@@ -1,30 +1,27 @@
-;; nnimap.el --- IMAP library for emacs
+;;; nnimap.el --- IMAP library for emacs
 
-;; Copyright (C) 1998 Simon Josefsson
-;; Copyright (C) 1998 Jim Radford
-;; Copyright (C) 1997 John McClary Prevost
+;;; Copyright (C) 1998 Simon Josefsson
+;;; Copyright (C) 1998 Jim Radford
+;;; Copyright (C) 1997 John McClary Prevost
 
-;; This file is part of Emacs.
+;;; This program is free software; you can redistribute it and/or modify
+;;; it under the terms of the GNU General Public License as published by
+;;; the Free Software Foundation; either version 2 of the License, or
+;;; (at your option) any later version.
 
-;; Emacs is free software; you can redistribute it and/or modify it
-;; under the terms of the GNU General Public License as published by
-;; the Free Software Foundation; either version 2, or (at your option)
-;; any later version.
+;;; This program is distributed in the hope that it will be useful,
+;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
+;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+;;; GNU General Public License for more details.
 
-;; Emacs is distributed in the hope that it will be useful, but
-;; WITHOUT ANY WARRANTY; without even the implied warranty of
-;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-;; General Public License for more details.
-
-;; You should have received a copy of the GNU General Public License
-;; along with XEmacs; see the file COPYING.  If not, write to the Free
-;; Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-;; 02111-1307, USA.
+;;; You should have received a copy of the GNU General Public License
+;;; along with this program; if not, write to the Free Software
+;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 ;;; Commentary:
 
 ;;; This file provides a low level interface to an IMAP server.
-;;; It provides functions to send commands an a wait for responses.
+;;; It provides functions to send commands and wait for responses.
 ;;; The results are parsed into lisp expressions and returned.
 ;;; IMAP untagged responses are stuffed in to obarrays based on
 ;;; the current group.  All the parameter variables become local
@@ -34,8 +31,6 @@
 ;;;   On expunge, remove messages from message-data. Note it doesn't
 ;;;     return UIDs.  Ouch.
 
-;;; $Id: imap4rev1.el,v 1.5.1.4.1.3 1998/06/21 23:13:31 jas Exp $
-
 (eval-and-compile
   (unless (fboundp 'open-network-stream)
     (require 'tcp)))
@@ -67,9 +62,6 @@ networked stream to the server.")
 (defvar imap-default-name nil
   "*Your name, should you choose to accept it.")
 
-(defvar imap-default-password nil
-  "*Your password.")
-
 (defvar imap-last-status nil
   "*Status returned by last IMAP command")
 
@@ -143,7 +135,6 @@ networked stream to the server.")
                       imap-do-login
                       imap-message-data
                       imap-default-name
-		      imap-default-password
                       imap-last-status
                       imap-process
                       imap-tag-num
@@ -155,9 +146,9 @@ any of these to `imap-open-server'.")
 
 
 ;; If non nil these hold the name of a buffer to put debug into
-(defvar imap-log   "*imap-log*")     ; imap session trace
-(defvar imap-last  "*imap-last*")    ; last line we attempted to parse
-(defvar imap-debug "*imap-debug*")   ; random debug spew
+(defvar imap-log   "*imap-log*")   ; imap session trace
+(defvar imap-last  "*imap-last*")  ; last line we attempted to parse
+(defvar imap-debug "*imap-debug*") ; random debug spew
 
 (when imap-debug ; (untrace-all)
   (require 'trace)
@@ -168,10 +159,12 @@ any of these to `imap-open-server'.")
 	  imap-select-mailbox
           imap-send-command
           imap-send-command-wait
+          imap-send-commands-wait
           imap-folder-set
           imap-folder-get
           imap-folder-plist
           imap-dispatch
+	  imap-login
           imap-message-set
           imap-message-get
           imap-message-map
@@ -189,6 +182,12 @@ any of these to `imap-open-server'.")
 
 ;;; Compatibility
 
+(or (fboundp 'char-int)
+    (fset 'char-int (symbol-function 'identity)))
+
+(or (fboundp 'int-char)
+    (fset 'int-char (symbol-function 'identity)))
+
 (if (not (fboundp 'remassoc))
     (defun remassoc (key alist)
       "Delete by side effect any elements of LIST whose car is `equal' to KEY.
@@ -202,6 +201,15 @@ the value of `foo'."
 	    (setcdr alist (remassoc key (cdr alist)))
 	    alist))))
 
+(if (not (fboundp 'save-current-buffer))
+    (defmacro save-current-buffer (&rest body)
+      "Save the current buffer; execute BODY; restore the current buffer.
+Executes BODY just like `progn'."
+      (` (let ((orig-buffer (current-buffer)))
+	   (unwind-protect
+	       (progn (,@ body))
+	     (set-buffer orig-buffer))))))
+
 (if (not (fboundp 'with-current-buffer))
     (defmacro with-current-buffer (buffer &rest body)
       "Execute the forms in BODY with BUFFER as the current buffer.
@@ -243,8 +251,7 @@ See also `plist-to-alist'."
 (defun imap-server-opened (&optional buffer)
   (with-current-buffer (or buffer (current-buffer))
     (and imap-process
-         (member (process-status imap-process) '(open run))
-	 (eq (car (imap-send-command-wait "NOOP")) 'OK))))
+         (member (process-status imap-process) '(open run)))))
 
 (defun imap-close-server (&optional buffer autologout)
   "Logout if needed and close down the process.  Clean out buffer.
@@ -289,7 +296,7 @@ values such that the buffer will be suitable for opening a new server."
 ;; have `imap-send-command-wait'ing commands pass 
 ;; `imap-cb-tag-default' itself.  Maybe `imap-wait-for-tag' should
 ;; be internal.  It currently doesn't accept a buffer (nor set one).
-;; I guees until then this is an internal command also.  Use
+;; I guess until then this is an internal command also.  Use
 ;; `imap-send-command-wait' instead.
 (defun imap-send-command (command &optional buffer callback)
   "Send a COMMAND to the server for BUFFER returning the command's TAG.  If
@@ -320,23 +327,23 @@ be `concat'ed together.  The buffers are sent as IMAP string literals."
                 strings)
           (setq imap-cb-tag-alist ;; removed by `imap-dispatch'
                 (cons (cons "+" 'imap-cb-tag-default) imap-cb-tag-alist)))
-        (setq strings (nreverse strings))
+        (setq strings (apply 'concat (nreverse strings)))
         (and imap-log (with-current-buffer (get-buffer-create imap-log)
                         (buffer-disable-undo)
                         (goto-char (point-max))
-                        (insert (apply 'concat strings))))
-        (process-send-string nil (apply 'concat strings)) 
+                        (insert strings)))
+        (process-send-string nil strings)
         (setq strings nil)
         (when commands ; buffer
           (if (not (imap-wait-for-tag "+"))
               (setq commands nil) ; Don't bother sending the rest
-            (and imap-log (with-current-buffer (get-buffer-create imap-log)
-                            (goto-char (point-max))
-                            (insert-buffer-substring (car commands))))
-            (let ((process imap-process))
-              (with-current-buffer (car commands)
-                (process-send-region process (point-min) (point-max))))
-            (pop commands))))
+	    (and imap-log (with-current-buffer (get-buffer-create imap-log)
+			    (goto-char (point-max))
+			    (insert-buffer-substring (car commands))))
+	    (let ((process imap-process))
+	      (with-current-buffer (car commands)
+		(process-send-region process (point-min) (point-max))))
+	    (pop commands))))
       tag)))
 
 (defun imap-send-command-wait (command &optional buffer)
@@ -380,21 +387,21 @@ in order.  See `imap-send-command-wait'."
                      imap-current-folder) imap-data-folder) prop)))
 
 (defun imap-login (&optional name buffer)
-  "Login to server, optionally using NAME."
+  "Login to server, optionally using NAME.   If imap-password is bound and
+non-nil, use it's content for password."
   (with-current-buffer (or buffer (current-buffer))
     (if imap-do-login
         (let* ((user (or name
                          imap-default-name
                          (read-from-minibuffer
                           (concat "IMAP Name for " imap-current-server ": "))))
-               (passwd (or imap-default-password
+               (passwd (or (and (boundp 'imap-password) imap-password)
 			   (funcall (or (and (load "passwd" t) 'read-passwd) 
 					(and (require 'ange-ftp) 'ange-ftp-read-passwd))
 				    (concat "IMAP Password for " user "@" 
 					    imap-current-server ": ")))))
-;          (prog1
-              (eq 'OK (car (imap-send-command-wait (concat "LOGIN " user " " passwd)))))
-;            (fillarray passwd 0)))
+	  (eq 'OK (car (imap-send-command-wait 
+			(concat "LOGIN " user " " passwd)))))
       t)))
 
 ;;; Internal functions
@@ -410,23 +417,36 @@ in order.  See `imap-send-command-wait'."
   (open-network-stream name buffer host port))
 
 (defun imap-open-imtest-stream (name buffer host &optional port)
-  (setq imap-eol "\n")
   (let ((process (start-process name (or buffer (current-buffer))
                                 "imtest" "-kp" host "imap")))
-    (when process
-      (message "Opening Kerberized IMAP connection...")
-      ;; Result of authentication is a string like __Full privacy protection__
-      (while (not (re-search-backward "__\\(.*\\)__\n" nil t nil
-                                      (process-buffer process)))
-        (accept-process-output process 5)) ; Yes, this is an oo loop
-      (message (concat "Opening Kerberized IMAP connection... "
-                       (buffer-string (match-beginning 1) (match-end 1))))
-      (delete-region (point-min) (match-end 0))
-      ;; If the __string__ contains "failed" authentication failed
-      ;; but thats ok, you can just login normal like
-      (setq imap-do-login (string-match "failed" (match-string 1)))
-      (message "Opening Kerberized IMAP connection...done"))
-    process))
+    (with-current-buffer (process-buffer process)
+      (setq imap-eol "\n")
+      (setq imap-do-login nil) ;; don't login even if kerberos auth fails
+      (when process
+	(message "Opening Kerberized IMAP connection...")
+	;; Result of authentication is a string: __Full privacy protection__
+	(while (not (or 
+		     ;; if everything is ok, this will match...
+		     (re-search-backward "__\\(.*\\)__\n" nil t)
+		     ;; ...errors will match this
+		     (re-search-backward 
+		      "\\. \\([^OA][^KU][^T][^H].*\\)\r\n" nil t)))
+	  (accept-process-output process 5)) ; Yes, this is an oo loop
+	(let ((response (match-string 1)))
+	  (erase-buffer)
+	  (message "Kerberized IMAP connection: %s" response)
+	  ;; If the __string__ contains "failed" authentication failed
+	  ;; (imtest will bug out if you try to login the usual way, so
+	  ;; close connection with an error)
+	  (when (string-match "failed\\|NO\\|BAD" response)
+	    (mapc 'make-variable-buffer-local imap-locals) ; just in case
+	    ;; XXX logout here (can't use send-command since we don't
+	    ;; have the server opened..)
+	    (delete-process imap-process)
+	    (mapc (lambda (local) (set local (default-value local))) imap-locals)
+	    (erase-buffer)
+	    (error "imtest: %s" response))))
+      process)))
 
 (defun imap-arrival-filter (proc string)
   "Process filter for imap process.  Stow the string, then call the routines
@@ -541,7 +561,7 @@ this returns nil."
 (defun imap-parse-line () 
 
   ;; Parse one server response.  We do translations to buffer so we
-  ;; can use `read'.  We quote the UIDVALIDITY number to avoind 28-bit
+  ;; can use `read'.  We quote the UIDVALIDITY number to avoid 28-bit
   ;; integer limitations.  Quotes [] into {} so that things like
   ;; BODY[HEADER] are one symbol, backslashify things, turn CRLF into
   ;; LF.  Call this at the start of the block to suck in, narrowed to
@@ -563,7 +583,7 @@ this returns nil."
         (narrow-to-region (point) (- (point-max) 2))
         (insert "\"")
         (while (re-search-forward "\\\\\\|\"" nil 00)
-          (replace-match "\\\\\&"))
+          (replace-match "\\\\\\&"))
         (insert "\""))
       (goto-char (point-min)))
     (goto-char (- (point-max) 2))
@@ -606,7 +626,7 @@ this returns nil."
                    (narrow-to-region (point) (+ (point) jump-amount))
                    (insert "\"")
                    (while (re-search-forward "\\\\\\|\"" nil 00)
-                     (replace-match "\\\\\&"))
+                     (replace-match "\\\\\\&"))
                    (goto-char (point-min))
                    (while (search-forward "\r\n" nil 00)
                      (replace-match "\n" nil t))
