@@ -89,7 +89,7 @@
 
 (nnoo-declare nnimap) ; we derive from no one
 
-(defvar nnimap-version "nnimap 0.3.9")
+(defvar nnimap-version "nnimap 0.3.10")
 
 (defvoo nnimap-list-pattern "*" 
 "*PATTERN or list of PATTERNS use to limit available groups.  
@@ -228,6 +228,7 @@ If slow, use the exact MESSAGES attribute.")
           nnimap-request-group
           nnimap-request-head
           nnimap-request-list
+          nnimap-request-list-mapper
           nnimap-request-newgroups
           nnimap-request-post
           nnimap-request-rename-group
@@ -369,7 +370,8 @@ If slow, use the exact MESSAGES attribute.")
     (with-current-buffer nnimap-server-buffer
       (when (stringp article)
         ;; Find the article number by Message-Id
-        (let ((articles (nnimap-search (concat "HEADER Message-Id " article))))
+        (let ((articles (nnimap-search 
+			 (format "HEADER Message-Id %d" article))))
           (when (cdr articles)
             (message 
              "IMAP Warning: Message Id search yielded more than one article %s"
@@ -401,12 +403,15 @@ If slow, use the exact MESSAGES attribute.")
 ;;; we can;t give a good answer for the lowest message number,
 ;;; so I think that request-update-info should somehow override
 ;;; this once it figures out what the real lowest is.
+;;;
+;;; The 'fast' concept is broken. Gnus =relies= on getting data
+;;; back from this function even if called with fast=t. Yuck.
 (deffoo nnimap-request-group (group &optional server fast)
   (when (nnimap-possibly-change-group group server) ;; SELECT group
-    (if fast
-	(with-current-buffer nntp-server-buffer 
-	  (erase-buffer) ;; gnus bug?
-	  t)
+;    (if fast
+;	(with-current-buffer nntp-server-buffer 
+;	  (erase-buffer)
+;	  t)
       (with-current-buffer nnimap-server-buffer
 	(let ((imap-message-data (make-vector 2 0))
 	      (exists (imap-folder-get 'EXISTS))
@@ -422,7 +427,7 @@ If slow, use the exact MESSAGES attribute.")
 	      (insert (format "211 %d %d %d %s\n" 
 			      exists (apply 'min articles)
 			      (apply 'max articles) group)))
-	    t))))))
+	    t)))));)
 
 (defun gnus-group-normally-qualified (backend server group)
   ;; This is the format for groups in the group-info.
@@ -508,8 +513,8 @@ If slow, use the exact MESSAGES attribute.")
   (when (nnimap-possibly-change-server server)
     (with-current-buffer nntp-server-buffer (erase-buffer))
     (with-current-buffer nnimap-server-buffer
-      (setq imap-data-folder
-            (make-vector imap-convenient-group-prime 0))
+      (setq imap-data-folder (make-vector imap-convenient-group-prime 0)
+	    imap-current-folder nil)
       (let* ((patterns (mapcar (lambda (p) (if (consp p) p (cons "\"\"" p)))
                                (if (or (atom nnimap-list-pattern)
                                        (atom (cdr nnimap-list-pattern)))
@@ -612,27 +617,29 @@ If slow, use the exact MESSAGES attribute.")
   'mail)
 
 ;; Move article ARTICLE from group GROUP on current server to group
-;; TO-GROUP.  Creates mailboxes if necessery. If TO-GROUP is nil, no
-;; actions are taken.
+;; TO-GROUP.  
 (defun nnimap-split-copy-delete-article (article group to-group)
-  (when to-group
-    (when (nnimap-ok-p (nnimap-send-command-wait 
-			(format "COPY %d %s" article to-group)))
-      (let ((store (nnimap-send-command-wait 
-		    (format "UID STORE %d +FLAGS (\\Seen \\Deleted)" 
-			    article))))
-	(if (nnimap-ok-p store)
-	    (message "IMAP split: moved %s:%d to %s" group article to-group)
-	  (error "IMAP flag store failed: %s (you may have unread mail marked as read!)" store))))))
+  (when (nnimap-ok-p (nnimap-send-command-wait  ;;  xxx UID COPY is not std
+		      (format "UID COPY %d %s" article to-group))) 
+    (let ((store 
+	   (nnimap-send-command-wait 
+	    (format "UID STORE %d +FLAGS (\\Flagged)" article))))
+      (if (nnimap-ok-p store)
+	  (message "IMAP split: moved %s:%d to %s" group article to-group)
+	(error "IMAP flag store failed: %s (you may have unread mail marked as read!)" store)))))
 
 (defun nnimap-split-move-article (article group to-group)
-  (unless (nnimap-split-copy-delete-article article group to-group)
-    (message "Could not find mailbox %s, creating..." to-group)
-    (if (nnimap-ok-p (nnimap-send-command-wait
-		      (format "CREATE %s" to-group)))
-	(nnimap-split-copy-delete-article article group to-group))))
+  (when to-group
+    (unless (nnimap-split-copy-delete-article article group to-group)
+      (message "Could not find mailbox %s, creating..." to-group)
+      (if (nnimap-ok-p (nnimap-send-command-wait
+			(format "CREATE %s" to-group)))
+	  (nnimap-split-copy-delete-article article group to-group)
+	(message "Could not create mailbox %s: %s" 
+		 to-group imap-last-status)))))
 
-(defun nnimap--split-move-article (article group to-group)
+;; doesn't use COPY but instead FETCH+APPEND. Slower.
+(defun nnimap-split--move-article (article group to-group)
   (when to-group
     (message "IMAP split: moving %s:%d to %s" group article to-group)
     
@@ -665,10 +672,10 @@ If slow, use the exact MESSAGES attribute.")
 	  ;; serious damage
 	  (error "Couldn't remove Seen flag on article %s:%d (you might have unread mail flagged as read!)" group article))))))
 
-;; tries to match all rules in RULES against content of nntp-server-buffer,
-;; returns a list of groups in RULES that matched. List will contain
-;; nil.
-(defun nnimap-split-to-groups (rules)
+;; tries to match all rules in nnimap-split-rule against content of 
+;; nntp-server-buffer, returns a list of groups that matched. List will 
+;; contain nil, which should be disregarded. (todo: fix nil's)
+(defun nnimap-split-to-groups ()
   (with-current-buffer nntp-server-buffer
     (mapcar (lambda (rule) 
 	      (let ((group (car rule))
@@ -676,7 +683,7 @@ If slow, use the exact MESSAGES attribute.")
 		(goto-char (point-min))
 		(when (re-search-forward regexp nil t)
 		  group)))
-	    rules)))
+	    (with-current-buffer nnimap-server-buffer nnimap-split-rule))))
 
 ;; 47
 (defun nnimap-split-articles (&optional group server)
@@ -693,12 +700,14 @@ If slow, use the exact MESSAGES attribute.")
 	    (while (setq article (pop unseens))
 	      (when (nnimap-request-head article)
 		;; article into what groups?
-		(let ((groups (nnimap-split-to-groups nnimap-split-rule)))
-		  ;; move it
+		(let ((groups (nnimap-split-to-groups)))
+		  ;; move it there
 		  (if nnimap-split-crosspost
+		      ;; move to all boxes
 		      (let (to-group)
 			(while (setq to-group (pop groups))
 			  (nnimap-split-move-article article inbox to-group)))
+		    ;; move to first non-nil box
 		    (let ((group (until (pop groups))))
 		      (nnimap-split-move-article article inbox 
 						 group))))))))))))
